/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.gui.FilmStrip;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.geom.RoundRectangle2D;
import java.awt.image.BufferedImage;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.EmptyBorder;

import ca.uvic.csr.shrimp.resource.ResourceHandler;
import ca.uvic.csr.shrimp.util.GradientPainter;


/**
 * 
 * 
 * @author Chris Callendar
 * @date 31-Oct-06
 */
public class SnapShotsPanel extends JPanel {

	private static final Color END_COLOR = new Color(54, 130, 191);
	private static final Color START_COLOR = new Color(0, 4, 24);
	private static final Color BORDER_COLOR = Color.black;
	private static final Color WHITE_BOX_COLOR = new Color(224, 224, 224);
	private static final Color BG = new Color(57, 131, 193);
	private static final Color LABEL_COLOR = new Color(192, 255, 255);
	private static final int BORDER_HEIGHT = 24;
	private static final int WHITE_BOX_SIZE = 10;
	private static final int LABEL_HEIGHT = 16;


	// these constants decide how many frames and how far apart to display them
	public static final int DIVIDER_WIDTH = 5;
	private static final int ARC = DIVIDER_WIDTH * 2;
	private static final int NUM_SNAP_SHOTS = 6;

	public static final int IMAGE_WIDTH = 100;
	public static final int IMAGE_HEIGHT = IMAGE_WIDTH;
	public static final int IMAGE_BG_WIDTH = IMAGE_WIDTH + (2 * DIVIDER_WIDTH);
	public static final int IMAGE_BG_HEIGHT = IMAGE_HEIGHT + (2 * DIVIDER_WIDTH);
	
	// Filmstrip overall height and width
	public static final int FILM_STRIP_WIDTH = ((IMAGE_BG_WIDTH + (2 * DIVIDER_WIDTH)) * NUM_SNAP_SHOTS);
	public static final int FILM_STRIP_HEIGHT = IMAGE_BG_HEIGHT + (2 * BORDER_HEIGHT) + (2 * DIVIDER_WIDTH) + LABEL_HEIGHT; 
	
	private JPanel mainPanel;
	
	private Vector snapShots;
	private SnapShotImagePanel[] imagePanels;
	private boolean hasChanged;
	private boolean locked = false;
	private final FilmStrip filmStrip;
	private Frame parent;
	
	public SnapShotsPanel(FilmStrip filmStrip, Frame parent) {
		super();
		this.filmStrip = filmStrip;
		this.snapShots = new Vector(NUM_SNAP_SHOTS);
		this.imagePanels = new SnapShotImagePanel[NUM_SNAP_SHOTS];
		this.hasChanged = false;
		this.parent = parent;
		
		initialize();
	}

	private void initialize() {
		setLayout(null);
		
		mainPanel = new JPanel(new GridLayout(1, NUM_SNAP_SHOTS, 0, 0));
		mainPanel.setBounds(0, 0, FILM_STRIP_WIDTH, FILM_STRIP_HEIGHT);
		mainPanel.setBorder(new FilmStripBorder());
		mainPanel.setBackground(BG);
		add(mainPanel, BorderLayout.NORTH);
		
		for (int i = 0; i < NUM_SNAP_SHOTS; i++) {
			SnapShotImagePanel pnl = new SnapShotImagePanel();
			mainPanel.add(pnl);
			imagePanels[i] = pnl;
		}
	}
	
	public boolean hasChanged() {
		return hasChanged;
	}

	public void setChanged(boolean b) {
		hasChanged = false;
	}
	
	public Vector getSnapShots() {
		return new Vector(snapShots);
	}
	
	public int getSnapShotCount() {
		return snapShots.size();
	}
	
	private void updateImagePanels() {
		if (!locked) {
			for (int i = 0; i < imagePanels.length; i++) {
				SnapShot snapShot = null;
				if (i < snapShots.size()) {
					snapShot = (SnapShot) snapShots.get(i);
				}
				imagePanels[i].setSnapShot(snapShot);
			}
			if (filmStrip != null) {
				filmStrip.snapShotsChanged();
			}
		}
	}

	/**
	 * Adds the given snapshot at the given position.
	 * @param snapShot the snapshot to add
	 * @param pos the position to add the snapshot at
	 */
	private void addSnapShot(SnapShot snapShot, int pos) {
		locked = true;
		snapShots.insertElementAt(snapShot, pos);
		while (snapShots.size() > NUM_SNAP_SHOTS) {
			removeSnapShot(snapShots.size() - 1);
		}
		locked = false;
		updateImagePanels();
		hasChanged = true;
	}
	
	/**
	 * Adds the given snapshot at the beginning of the filmstrip
	 * @param snapShot the snapshot to add
	 */
	public void addSnapShot(SnapShot snapShot) {
		addSnapShot(snapShot, 0);
	}
	
	public void addSnapShots(Vector snapShotsToAdd) {
		locked = true;
		// add all the snapshots to the front in reverse order
		for (int i = snapShotsToAdd.size() - 1; i >= 0; i--) {
			SnapShot snapShot = (SnapShot) snapShotsToAdd.get(i);
			snapShots.insertElementAt(snapShot, 0);
		}
		// remove any extras
		while (snapShots.size() > NUM_SNAP_SHOTS) {
			removeSnapShot(snapShots.size() - 1);
		}
		locked = false;
		// update the panels
		updateImagePanels();
		hasChanged = true;
	}
	
	/**
	 * Removes all the snapshots in the filmstrip
	 */
	public void removeAllSnapShots() {
		locked = true;
		while (snapShots.size() > 0) {
			removeSnapShot(0);
		}
		locked = false;
		updateImagePanels();
	}


	/**
	 * Removes the given snapshot.
	 */
	private void removeSnapShot(SnapShot snapShot) {
		if (snapShot != null) {
			int index = snapShots.indexOf(snapShot);
			if (index != -1) {
				removeSnapShot(index);
			} else {
				System.err.println("Warning - no snap shot found to remove");
			}
		}
	}
	
	/**
	 * Removes the snapshot at the given position
	 * @param pos the position of the snapshot to remove
	 */
	private void removeSnapShot(int pos) {
		SnapShot snapShot = (SnapShot) snapShots.elementAt(pos);
		snapShot.cleanUp();
		snapShots.removeElementAt(pos);
		hasChanged = true; // update the flag
		updateImagePanels();
	}
	
	public static void main(String[] args) {
		JFrame frame = new JFrame("FilmStrip");
		frame.setIconImage(ResourceHandler.getIconAsImage("icon_filmstrip.gif"));
		final Dimension dim = new Dimension(FILM_STRIP_WIDTH + 15, FILM_STRIP_HEIGHT + 40);
		frame.setPreferredSize(dim);
		frame.setSize(dim);
		frame.setAlwaysOnTop(true);
		frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		frame.getContentPane().setLayout(new BorderLayout());
		final SnapShotsPanel snapShotsPanel = new SnapShotsPanel(null, frame);
		// create a white image for testing the buttons 
		final BufferedImage img = new BufferedImage(IMAGE_WIDTH, IMAGE_HEIGHT, BufferedImage.TYPE_INT_RGB);
		img.getGraphics().setColor(Color.white);
		img.getGraphics().fillRect(0, 0, IMAGE_WIDTH, IMAGE_HEIGHT);
		snapShotsPanel.imagePanels[0].image = img;
		snapShotsPanel.imagePanels[0].updateVisibility();
		frame.getContentPane().add(snapShotsPanel, BorderLayout.CENTER);
		frame.addWindowListener(new WindowAdapter() {
			public void windowClosed(WindowEvent e) {
				System.exit(0);
			}
		});
		frame.setVisible(true);
	}
	
	

	/** 
	 * Creates the filmstrip ticker-tape border.
	 * 
	 * @author Chris Callendar
	 * @date 1-Nov-06
	 */
	class FilmStripBorder extends EmptyBorder {

		private final int WHITE_BOX_SIZE_HALF = (WHITE_BOX_SIZE / 2);
		private final int MID = (BORDER_HEIGHT / 2) - WHITE_BOX_SIZE_HALF;

		public FilmStripBorder() {
			super(BORDER_HEIGHT, 0, BORDER_HEIGHT, 0);
		}
		
		public void paintBorder(Component c, Graphics g, int x, int y, int width, int height) {
			// draw a 1 pixel black border around the whole filmstrip
			g.setColor(BORDER_COLOR);
			g.drawRect(x, y, width - 1, height - 1);
			
			// paint the top strip and white boxes
			g.fillRect(x, y, width, BORDER_HEIGHT);
			paintWhiteBoxes(g, x + WHITE_BOX_SIZE_HALF, y + MID, WHITE_BOX_SIZE, x + width);
			
			// paint the bottom strip and white boxes
			g.setColor(BORDER_COLOR);
			g.fillRect(x, y + height - BORDER_HEIGHT, width, BORDER_HEIGHT);
			paintWhiteBoxes(g, x + WHITE_BOX_SIZE_HALF, y + height - BORDER_HEIGHT + MID, WHITE_BOX_SIZE, x + width);
		}
		
		private void paintWhiteBoxes(Graphics g, int startX, int y, int w, int maxX) {
			g.setColor(WHITE_BOX_COLOR);
			int x = startX;
			while (x < maxX) {
				g.fillRect(x, y, w, w);
				x += (2 * w);
			}
		}
		
	}
	
	/**
	 * 
	 * 
	 * @author Chris Callendar
	 * @date 1-Nov-06
	 */
	class SnapShotImagePanel extends JPanel implements ActionListener {

		private SnapShot snapShot;
		private Image image;
		private JButton btnRun;
		private JButton btnEdit;
		private JButton btnClose;
		private JLabel lblComment;
		private GradientPainter painter;
		
		public SnapShotImagePanel() {
			this.painter = new GradientPainter(START_COLOR, END_COLOR);
			this.painter.setAntiAliasing(true);
			this.painter.setBorderPainted(false);
			initialize();
			updateVisibility();
		}
		
		private void initialize() {
			setBorder(BorderFactory.createEmptyBorder(DIVIDER_WIDTH, DIVIDER_WIDTH, DIVIDER_WIDTH, DIVIDER_WIDTH));
			setLayout(new BorderLayout(0, 0));
			setOpaque(false);
			Dimension dim = new Dimension(IMAGE_BG_WIDTH + (DIVIDER_WIDTH * 2), IMAGE_BG_HEIGHT + (DIVIDER_WIDTH * 2) + LABEL_HEIGHT);
			setPreferredSize(dim);
			setMinimumSize(dim);
			setMaximumSize(dim);
			add(createTopPanel(), BorderLayout.NORTH);
			add(Box.createVerticalGlue(), BorderLayout.CENTER);
			add(createBottomLabel(), BorderLayout.SOUTH);
			addMouseListener(new MouseAdapter() {
				public void mouseClicked(MouseEvent e) {
					if ((e.getClickCount() >= 2) && (snapShot != null)) {
						btnRun.doClick();
					}
				}
			});
		}

		private JLabel createBottomLabel() {
			lblComment = new JLabel("", JLabel.CENTER);
			lblComment.setForeground(LABEL_COLOR);
			Font font = lblComment.getFont();
			lblComment.setFont(font.deriveFont(font.getSize2D() - 1f));
			lblComment.setBorder(BorderFactory.createCompoundBorder(
					BorderFactory.createMatteBorder(1, 1, 0, 0, Color.gray),
					BorderFactory.createMatteBorder(0, 0, 1, 1, Color.lightGray)));
			Dimension dim = new Dimension(IMAGE_WIDTH, 16);
			lblComment.setMinimumSize(dim);
			lblComment.setPreferredSize(dim);
			lblComment.setMaximumSize(dim);
			return lblComment;
		}

		private JPanel createTopPanel() {
			JPanel pnl = new JPanel(new FlowLayout(FlowLayout.RIGHT, 0, DIVIDER_WIDTH));	// right align buttons
			pnl.setOpaque(false);
			// add the edit/close button only if an image is present
			btnRun = createButton("icon_forward_small.gif", "Load this snapshot", "Run");
			btnEdit = createButton("icon_edit.gif", "Annotate this snapshot", "Edit");
			btnClose = createButton("icon_close_gray.gif", "icon_close_red.gif", "Delete this snapshot", "Close");

			pnl.add(btnRun);
			pnl.add(Box.createHorizontalStrut(1));
			pnl.add(btnEdit);
			pnl.add(Box.createHorizontalStrut(1));
			pnl.add(btnClose);
			pnl.add(Box.createHorizontalStrut(DIVIDER_WIDTH));	// to align horizontally with the image
			return pnl;
		}

		private JButton createButton(String iconPath, String rolloverIconPath, String tooltip, String actionCmd) {
			JButton btn = createButton(iconPath, tooltip, actionCmd);
			btn.setRolloverEnabled(true);
			btn.setRolloverIcon(ResourceHandler.getIcon(rolloverIconPath));
			return btn;
		}

		private JButton createButton(String iconPath, String tooltip, String actionCmd) {
			JButton btn = new JButton(ResourceHandler.getIcon(iconPath));
			btn.setToolTipText(tooltip);
			btn.setActionCommand(actionCmd);
			btn.addActionListener(this);
			Dimension size = new Dimension(18, 18);
			btn.setPreferredSize(size);
			btn.setMaximumSize(size);
			btn.setMinimumSize(size);
			btn.setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
			return btn;
		}
		
		public void actionPerformed(ActionEvent e) {
			String cmd = e.getActionCommand();
			if ("Run".equals(cmd)) {
				filmStrip.applyLayout(snapShot);
			} else if ("Edit".equals(cmd) && (snapShot != null)) {
				snapShot.promptUserToChangeComment(parent);
				setToolTipText(snapShot.getComment());
			} else if ("Close".equals(cmd)) {
				SnapShot old = snapShot;
				setSnapShot(null);	// make sure the image is removed and buttons hidden
				removeSnapShot(old);
			}
		}
		
		public void setSnapShot(SnapShot snapShot) {
			this.snapShot = snapShot;
			image = (snapShot != null ? snapShot.getPreviewShot() : null);
			updateVisibility();
		}
		
		private void updateVisibility() {
			String txt = "";
			txt = (snapShot != null ? snapShot.getComment() : "");
			txt = (txt == null ? "" : txt);
			lblComment.setVisible(txt.length() > 0);
			lblComment.setText(txt);
			lblComment.setToolTipText(txt);
			boolean vis = (image != null);
			btnRun.setVisible(vis);
			btnEdit.setVisible(vis);
			btnClose.setVisible(vis);
			repaint();
		}
		
		/**
		 * Draws the image if one exists.
		 */
		public void paintComponent(Graphics g) {
			super.paintComponent(g);
			// add a bit of padding 
			int x = DIVIDER_WIDTH;
			int y = DIVIDER_WIDTH; 
			paintBackground(g, x, y);
			// paint image after the background paint and center in the round rectangle
			if (image != null) {
				x += DIVIDER_WIDTH;
				y += DIVIDER_WIDTH; 
				g.setColor(Color.white);
				g.fillRect(x, y, IMAGE_WIDTH, IMAGE_HEIGHT);
				g.drawImage(image, x, y, this);
			}			
		}
		
		/**
		 * Paints the background gradient.
		 */
		private void paintBackground(Graphics g, int startX, int startY) {
			painter.paint(g, new RoundRectangle2D.Float(startX, startY, IMAGE_BG_WIDTH, IMAGE_BG_HEIGHT, ARC, ARC));
		}

	}
	
}
